#ifndef _ROW_ACL_H_

#include <postgres.h>
#include <fmgr.h>

typedef uint16 row_acl;

#define ROW_ACL_LENGTH		sizeof(row_acl)
#define ROW_ACL_COUNT		(ROW_ACL_LENGTH * 8 - 1)
#define ROW_ACL_BIT_LENGTH  ROW_ACL_COUNT
#define ROW_ACL_STR_LENGTH	(ROW_ACL_COUNT + 1)

#define ROW_ACL_USER_READ	16384	/* 10000 00000 00000 */
#define ROW_ACL_USER_WRITE	8192	/* 01000 00000 00000 */
#define ROW_ACL_USER_DELETE	4096	/* 00100 00000 00000 */
#define ROW_ACL_USER_INSERT	2048	/* 00010 00000 00000 */
#define ROW_ACL_USER_EXEC	1024	/* 00001 00000 00000 */
#define ROW_ACL_USER_ALL	31744	/* 11111 00000 00000 */

#define ROW_ACL_GROUP_READ	512	/* 00000 10000 00000 */
#define ROW_ACL_GROUP_WRITE	256	/* 00000 01000 00000 */
#define ROW_ACL_GROUP_DELETE	128	/* 00000 00100 00000 */
#define ROW_ACL_GROUP_INSERT	64	/* 00000 00010 00000 */
#define ROW_ACL_GROUP_EXEC	32	/* 00000 00001 00000 */
#define ROW_ACL_GROUP_ALL	992	/* 00000 11111 00000 */

#define ROW_ACL_OTHER_READ	16	/* 00000 00000 10000 */
#define ROW_ACL_OTHER_WRITE	8	/* 00000 00000 01000 */
#define ROW_ACL_OTHER_DELETE	4	/* 00000 00000 00100 */
#define ROW_ACL_OTHER_INSERT	2	/* 00000 00000 00010 */
#define ROW_ACL_OTHER_EXEC	1	/* 00000 00000 00001 */
#define ROW_ACL_OTHER_ALL	31	/* 00000 00000 11111 */

#define ROW_ACL_NOTHING		0
#define ROW_ACL_ALL		((1 << ROW_ACL_COUNT) - 1)

#define ROW_SHL_USR		10
#define ROW_SHL_GRP		5

#define ROW_USER_NOAUTH		0
#define ROW_USER_ROOT		1

/* ------------------------------------------------------------- */

/* get acl arg by value */
#define ROW_ACL_GETARG(_p_index)   \
    PG_GETARG_UINT16(_p_index)

/* return acl by value */
#define ROW_ACL_RETURN(_p_rowacl)   \
    return UInt16GetDatum(_p_rowacl)

/* datum convertion */
#define ROW_ACL_TO_DATUM(_p_acl)    \
    UInt16GetDatum(_p_acl)

#define ROW_ACL_FROM_DATUM(_p_dtm)  \
    DatumGetUInt16(_p_dtm)

#define ROW_ACL_BITS_MASK   0x1f
#define ROW_ACL_STATE_COUNT (1 <<  5)   /* number of states for 5 bit field */
#define ROW_ACL_TOTAL_COUNT (1 << 15)   /* number of states for whole acl */

/* access to group of bits */
#define ROW_ACL_USR_BITS(_p_acl)    \
    (((_p_acl) >> ROW_ACL_USR_SHIFT) & ROW_ACL_BITS_MASK)

#define ROW_ACL_GRP_BITS(_p_acl)    \
    (((_p_acl) >> ROW_ACL_GRP_SHIFT) & ROW_ACL_BITS_MASK)

#define ROW_ACL_OTH_BITS(_p_acl)    \
    (((_p_acl) >> ROW_ACL_OTH_SHIFT) & ROW_ACL_BITS_MASK)

/* access to one bit of permissions */
#define ROW_ACL_R_BIT(_p_acl)   \
    (((_p_acl) >> ROW_ACL_R_SHIFT) & 1)

#define ROW_ACL_W_BIT(_p_acl)   \
    (((_p_acl) >> ROW_ACL_W_SHIFT) & 1)

#define ROW_ACL_D_BIT(_p_acl)   \
    (((_p_acl) >> ROW_ACL_D_SHIFT) & 1)

#define ROW_ACL_I_BIT(_p_acl)   \
    (((_p_acl) >> ROW_ACL_I_SHIFT) & 1)

#define ROW_ACL_X_BIT(_p_acl)   \
    (((_p_acl) >> ROW_ACL_X_SHIFT) & 1)

#define ROW_ACL_USR_SHIFT   ROW_SHL_USR
#define ROW_ACL_GRP_SHIFT   ROW_SHL_GRP
#define ROW_ACL_OTH_SHIFT   0

#define ROW_ACL_R_SHIFT     4
#define ROW_ACL_W_SHIFT     3
#define ROW_ACL_D_SHIFT     2
#define ROW_ACL_I_SHIFT     1
#define ROW_ACL_X_SHIFT     0

#define ROW_ACL_USR_R_SHIFT (ROW_ACL_USR_SHIFT + ROW_ACL_R_SHIFT)
#define ROW_ACL_GRP_R_SHIFT (ROW_ACL_GRP_SHIFT + ROW_ACL_R_SHIFT)
#define ROW_ACL_OTH_R_SHIFT (ROW_ACL_OTH_SHIFT + ROW_ACL_R_SHIFT)

#define ROW_ACL_USR_W_SHIFT (ROW_ACL_USR_SHIFT + ROW_ACL_W_SHIFT)
#define ROW_ACL_GRP_W_SHIFT (ROW_ACL_GRP_SHIFT + ROW_ACL_W_SHIFT)
#define ROW_ACL_OTH_W_SHIFT (ROW_ACL_OTH_SHIFT + ROW_ACL_W_SHIFT)

#define ROW_ACL_USR_D_SHIFT (ROW_ACL_USR_SHIFT + ROW_ACL_D_SHIFT)
#define ROW_ACL_GRP_D_SHIFT (ROW_ACL_GRP_SHIFT + ROW_ACL_D_SHIFT)
#define ROW_ACL_OTH_D_SHIFT (ROW_ACL_OTH_SHIFT + ROW_ACL_D_SHIFT)

#define ROW_ACL_USR_I_SHIFT (ROW_ACL_USR_SHIFT + ROW_ACL_I_SHIFT)
#define ROW_ACL_GRP_I_SHIFT (ROW_ACL_GRP_SHIFT + ROW_ACL_I_SHIFT)
#define ROW_ACL_OTH_I_SHIFT (ROW_ACL_OTH_SHIFT + ROW_ACL_I_SHIFT)

#define ROW_ACL_USR_X_SHIFT (ROW_ACL_USR_SHIFT + ROW_ACL_X_SHIFT)
#define ROW_ACL_GRP_X_SHIFT (ROW_ACL_GRP_SHIFT + ROW_ACL_X_SHIFT)
#define ROW_ACL_OTH_X_SHIFT (ROW_ACL_OTH_SHIFT + ROW_ACL_X_SHIFT)

#define _ROW_ACL_H_
#endif

