/* -*- C -*-
 * File: "/home/arn/proj/cms1/db/row_acl/row_acl_text.c"
 * Author: "Anton Yabchinskiy, <arn@antora.ru>"
 * Copyright: ""
 * Created: "Fri, 13 Jul 2007 09:15:00 +0400 (arn@doriath.my.domain)"
 * Last modified: "Tue, 14 Aug 2007 10:21:13 +0400 (arn@doriath.my.domain)"
 * vim:set sts=4 ts=4 sw=4 ft=c:
 * $Id: row_acl_text.c,v 1.5 2007/08/14 06:39:14 anton Exp $
 */

#include "row_acl_text.h"

char*
row_acl_bits_to_str(row_acl p_acl, char* p_buf)
{
    static const char   chrs[2][5] = {
        { '-', '-', '-', '-', '-' },
        { 'r', 'w', 'd', 'i', 'x' }
    };

    p_buf[0] = chrs[ROW_ACL_R_BIT(p_acl)][0];
    p_buf[1] = chrs[ROW_ACL_W_BIT(p_acl)][1];
    p_buf[2] = chrs[ROW_ACL_D_BIT(p_acl)][2];
    p_buf[3] = chrs[ROW_ACL_I_BIT(p_acl)][3];
    p_buf[4] = chrs[ROW_ACL_X_BIT(p_acl)][4];
    p_buf[5] = '\0';

    return p_buf;
}

char*
row_acl_to_str(row_acl p_acl, char* p_buf)
{
    row_acl_bits_to_str(ROW_ACL_USR_BITS(p_acl), p_buf +  0);
    row_acl_bits_to_str(ROW_ACL_GRP_BITS(p_acl), p_buf +  5);
    row_acl_bits_to_str(ROW_ACL_OTH_BITS(p_acl), p_buf + 10);

    return p_buf;
}

row_acl
row_acl_bits_from_str(const char* p_input, int* p_error)
{
    *p_error |= (p_input[0] != 'r' && p_input[0] != '-') |
                (p_input[1] != 'w' && p_input[1] != '-') |
                (p_input[2] != 'd' && p_input[2] != '-') |
                (p_input[3] != 'i' && p_input[3] != '-') |
                (p_input[4] != 'x' && p_input[4] != '-');

    return ((p_input[0] == 'r') << ROW_ACL_R_SHIFT) |
           ((p_input[1] == 'w') << ROW_ACL_W_SHIFT) |
           ((p_input[2] == 'd') << ROW_ACL_D_SHIFT) |
           ((p_input[3] == 'i') << ROW_ACL_I_SHIFT) |
           ((p_input[4] == 'x') << ROW_ACL_X_SHIFT);
}

row_acl
row_acl_from_abs_str(const char* p_input, int* p_error)
{
    return (row_acl_bits_from_str(p_input +  0, p_error) << ROW_ACL_USR_SHIFT) |
           (row_acl_bits_from_str(p_input +  5, p_error) << ROW_ACL_GRP_SHIFT) |
           (row_acl_bits_from_str(p_input + 10, p_error) << ROW_ACL_OTH_SHIFT);
}

row_acl
row_acl_from_sym_dat(row_acl p_result, const char* p_input, size_t p_size, int* p_error)
{
#if 0

    /*
     * weird, non-standard finite state automaton
     * for symbolic modes parsing. Uses functions
     * as quantities.
     */
    struct state_s {
        int     who_u;
        int     who_g;
        int     who_o;
        int     grant;
        row_acl result;
        int     is_fin;
    };

    typedef struct state_s  state_t;
    typedef void* (*quant_t)(char, state_t*);

    /* quantities prototypes */
    static void* q0(char, state_t*);
    static void* q1(char, state_t*);
    static void* q2(char, state_t*);
    static void* q3(char, state_t*);
    static void* q4(char, state_t*);

    /* quantities */
    void*
    q4(char p_c, state_t* p_s)
    {
        p_s->is_fin = 1;

        switch(p_c){
            case ' ' :
            case '\t' :
                return q4;
            case ',' :
                return q0;
        };
 
        return NULL;
    }

    void*
    q3(char p_c, state_t* p_s)
    {
        p_s->is_fin = 1;

        switch(p_c){
            case 'r' :
                if(p_s->grant){
                    p_s->result |=  (p_s->who_u << (ROW_ACL_USR_R_SHIFT));
                    p_s->result |=  (p_s->who_g << (ROW_ACL_GRP_R_SHIFT));
                    p_s->result |=  (p_s->who_o << (ROW_ACL_OTH_R_SHIFT));
                } else {
                    p_s->result &= ~(p_s->who_u << (ROW_ACL_USR_R_SHIFT));
                    p_s->result &= ~(p_s->who_g << (ROW_ACL_GRP_R_SHIFT));
                    p_s->result &= ~(p_s->who_o << (ROW_ACL_OTH_R_SHIFT));
                };
                return q3;
            case 'w' :
                if(p_s->grant){
                    p_s->result |=  (p_s->who_u << (ROW_ACL_USR_W_SHIFT));
                    p_s->result |=  (p_s->who_g << (ROW_ACL_GRP_W_SHIFT));
                    p_s->result |=  (p_s->who_o << (ROW_ACL_OTH_W_SHIFT));
                } else {
                    p_s->result &= ~(p_s->who_u << (ROW_ACL_USR_W_SHIFT));
                    p_s->result &= ~(p_s->who_g << (ROW_ACL_GRP_W_SHIFT));
                    p_s->result &= ~(p_s->who_o << (ROW_ACL_OTH_W_SHIFT));
                };
                return q3;
            case 'd' :
                if(p_s->grant){
                    p_s->result |=  (p_s->who_u << (ROW_ACL_USR_D_SHIFT));
                    p_s->result |=  (p_s->who_g << (ROW_ACL_GRP_D_SHIFT));
                    p_s->result |=  (p_s->who_o << (ROW_ACL_OTH_D_SHIFT));
                } else {
                    p_s->result &= ~(p_s->who_u << (ROW_ACL_USR_D_SHIFT));
                    p_s->result &= ~(p_s->who_g << (ROW_ACL_GRP_D_SHIFT));
                    p_s->result &= ~(p_s->who_o << (ROW_ACL_OTH_D_SHIFT));
                };
                return q3;
            case 'i' :
                if(p_s->grant){
                    p_s->result |=  (p_s->who_u << (ROW_ACL_USR_I_SHIFT));
                    p_s->result |=  (p_s->who_g << (ROW_ACL_GRP_I_SHIFT));
                    p_s->result |=  (p_s->who_o << (ROW_ACL_OTH_I_SHIFT));
                } else {
                    p_s->result &= ~(p_s->who_u << (ROW_ACL_USR_I_SHIFT));
                    p_s->result &= ~(p_s->who_g << (ROW_ACL_GRP_I_SHIFT));
                    p_s->result &= ~(p_s->who_o << (ROW_ACL_OTH_I_SHIFT));
                };
                return q3;
            case 'x' :
                if(p_s->grant){
                    p_s->result |=  (p_s->who_u << (ROW_ACL_USR_X_SHIFT));
                    p_s->result |=  (p_s->who_g << (ROW_ACL_GRP_X_SHIFT));
                    p_s->result |=  (p_s->who_o << (ROW_ACL_OTH_X_SHIFT));
                } else {
                    p_s->result &= ~(p_s->who_u << (ROW_ACL_USR_X_SHIFT));
                    p_s->result &= ~(p_s->who_g << (ROW_ACL_GRP_X_SHIFT));
                    p_s->result &= ~(p_s->who_o << (ROW_ACL_OTH_X_SHIFT));
                };
                return q3;
            case 'a' :
                if(p_s->grant){
                    p_s->result |=  (p_s->who_u ? ROW_ACL_USER_ALL  : 0);
                    p_s->result |=  (p_s->who_g ? ROW_ACL_GROUP_ALL : 0);
                    p_s->result |=  (p_s->who_o ? ROW_ACL_OTHER_ALL : 0);
                } else {
                    p_s->result &= ~(p_s->who_u ? ROW_ACL_USER_ALL  : 0);
                    p_s->result &= ~(p_s->who_g ? ROW_ACL_GROUP_ALL : 0);
                    p_s->result &= ~(p_s->who_o ? ROW_ACL_OTHER_ALL : 0);
                };
                return q3;
            case ',' :
                p_s->who_u =
                p_s->who_g =
                p_s->who_o = 0;
                p_s->grant = 0;
                return q0;
            case ' ' :
            case '\t' :
                p_s->who_u =
                p_s->who_g =
                p_s->who_o = 0;
                p_s->grant = 0;
                return q4;
        };

        return NULL;
    }

    void*
    q2(char p_c, state_t* p_s)
    {
        p_s->is_fin = 0;

        switch(p_c){
            case 'r' :
            case 'w' :
            case 'd' :
            case 'i' :
            case 'x' :
            case 'a' :
                return q3(p_c, p_s);
        };
 
        return NULL;
    }

    void*
    q1(char p_c, state_t* p_s)
    {
        p_s->is_fin = 0;

        switch(p_c){
            case 'u' :
                p_s->who_u = 1;
                return q1;
            case 'g' :
                p_s->who_g = 1;
                return q1;
            case 'o' :
                p_s->who_o = 1;
                return q1;
            case 'a' :
                p_s->who_u = 1;
                p_s->who_g = 1;
                p_s->who_o = 1;
                return q1;
            case '+' :
                p_s->grant = 1;
                return q2;
            case '-' :
                p_s->grant = 0;
                return q2;
        };

        return NULL;
    }

    void*
    q0(char p_c, state_t* p_s)
    {
        p_s->is_fin = 0;

        switch(p_c){
            case ' ' :
            case '\t' :
                return q0;
            case 'u' :
            case 'g' :
            case 'o' :
            case 'a' :
                return q1(p_c, p_s);
        };

        return NULL;
    }

    /* data */
    quant_t quant = q0;
    state_t state = {
        .who_u = 0,
        .who_g = 0,
        .who_o = 0,
        .grant = 0,
        .result = p_result,
        .is_fin = 0
    };

    /* parsing */
    for(; *p_input != '\0' && quant != NULL && p_size > 0; ++p_input, --p_size){
        quant = quant(*p_input, &state);
    };

    /* error reporting */
    *p_error |= (quant == NULL) || (! state.is_fin);

    return state.result;

#else

    typedef enum {
        Q_INIT = 0x80,
        Q_FINI = 0x40,
        Q_STOP = 0x20,

        Q_EXPECT_WHO    = 0x0 | Q_INIT,
        Q_EXPECT_HOW    = 0x1,
        Q_EXPECT_PERMS  = 0x2,
        Q_EXPECT_WHAT   = 0x3 | Q_FINI,
        Q_EXPECT_END    = 0x4 | Q_FINI
    } state_t;

    state_t state = Q_EXPECT_WHO;
    int     who_u = 0;
    int     who_g = 0;
    int     who_o = 0;
    int     grant = 0;

    for(; *p_input != '\0' && state != Q_STOP && p_size > 0; ++p_input, --p_size){
        switch(state){
            /* eat spaces before 'who' section */
            case Q_EXPECT_WHO :
                switch(*p_input){
                    case ' ' :
                    case '\t' :
                        break;  /* leave in this state and keep eating spaces */

                    case 'u' :  /* seems like 'who' section begins ... */
                    case 'g' :
                    case 'o' :
                    case 'a' :
                        --p_input, ++p_size;    /* ... so, push char back ... */
                        state = Q_EXPECT_HOW;   /* ... and switch to next state */
                        break;

                    default :   /* unexpected char */
                        state = Q_STOP;
                        break;
                };
                break;

            /* interpret 'who' section */
            case Q_EXPECT_HOW :
                switch(*p_input){
                    case 'u' :
                        who_u = 1;
                        break;

                    case 'g' :
                        who_g = 1;
                        break;

                    case 'o' :
                        who_o = 1;
                        break;

                    case 'a' :
                        who_u =
                        who_g =
                        who_o = 1;
                        break;

                    case '+' :  
                        grant = 1;              /* will grant perms */
                        state = Q_EXPECT_PERMS; /* switch to next state */
                        break;

                    case '-' :
                        grant = 0;              /* will revoke perms */
                        state = Q_EXPECT_PERMS; /* switch to next state */
                        break;

                    default :   /* unexpected char */
                        state = Q_STOP;
                        break;
                };
                break;

            /* expect begin of 'perms' section */
            case Q_EXPECT_PERMS :
                switch(*p_input){
                    case 'r' :
                    case 'w' :
                    case 'd' :
                    case 'i' :
                    case 'x' :
                    case 'a' :
                        --p_input, ++p_size;    /* push back. Will be handled by next state */
                        state = Q_EXPECT_WHAT;  /* switch to next */
                        break;

                    default :   /* unexpected char */
                        state = Q_STOP;
                        break;
                };
                break;

            /* interpret grant/revoke section */
            case Q_EXPECT_WHAT :
                switch(*p_input){
                    case 'r' :
                        if(grant){
                            if(who_u) p_result |= ROW_ACL_USER_READ;
                            if(who_g) p_result |= ROW_ACL_GROUP_READ;
                            if(who_o) p_result |= ROW_ACL_OTHER_READ;
                        } else {
                            if(who_u) p_result &= ~ROW_ACL_USER_READ;
                            if(who_g) p_result &= ~ROW_ACL_GROUP_READ;
                            if(who_o) p_result &= ~ROW_ACL_OTHER_READ;
                        };
                        break;

                    case 'w' :
                        if(grant){
                            if(who_u) p_result |= ROW_ACL_USER_WRITE;
                            if(who_g) p_result |= ROW_ACL_GROUP_WRITE;
                            if(who_o) p_result |= ROW_ACL_OTHER_WRITE;
                        } else {
                            if(who_u) p_result &= ~ROW_ACL_USER_WRITE;
                            if(who_g) p_result &= ~ROW_ACL_GROUP_WRITE;
                            if(who_o) p_result &= ~ROW_ACL_OTHER_WRITE;
                        };
                        break;

                    case 'd' :
                        if(grant){
                            if(who_u) p_result |= ROW_ACL_USER_DELETE;
                            if(who_g) p_result |= ROW_ACL_GROUP_DELETE;
                            if(who_o) p_result |= ROW_ACL_OTHER_DELETE;
                        } else {
                            if(who_u) p_result &= ~ROW_ACL_USER_DELETE;
                            if(who_g) p_result &= ~ROW_ACL_GROUP_DELETE;
                            if(who_o) p_result &= ~ROW_ACL_OTHER_DELETE;
                        };
                        break;

                    case 'i' :
                        if(grant){
                            if(who_u) p_result |= ROW_ACL_USER_INSERT;
                            if(who_g) p_result |= ROW_ACL_GROUP_INSERT;
                            if(who_o) p_result |= ROW_ACL_OTHER_INSERT;
                        } else {
                            if(who_u) p_result &= ~ROW_ACL_USER_INSERT;
                            if(who_g) p_result &= ~ROW_ACL_GROUP_INSERT;
                            if(who_o) p_result &= ~ROW_ACL_OTHER_INSERT;
                        };
                        break;

                    case 'x' :
                        if(grant){
                            if(who_u) p_result |= ROW_ACL_USER_EXEC;
                            if(who_g) p_result |= ROW_ACL_GROUP_EXEC;
                            if(who_o) p_result |= ROW_ACL_OTHER_EXEC;
                        } else {
                            if(who_u) p_result &= ~ROW_ACL_USER_EXEC;
                            if(who_g) p_result &= ~ROW_ACL_GROUP_EXEC;
                            if(who_o) p_result &= ~ROW_ACL_OTHER_EXEC;
                        };
                        break;

                    case 'a' :
                        if(grant){
                            if(who_u) p_result |= ROW_ACL_USER_ALL;
                            if(who_g) p_result |= ROW_ACL_GROUP_ALL;
                            if(who_o) p_result |= ROW_ACL_OTHER_ALL;
                        } else {
                            if(who_u) p_result &= ~ROW_ACL_USER_ALL;
                            if(who_g) p_result &= ~ROW_ACL_GROUP_ALL;
                            if(who_o) p_result &= ~ROW_ACL_OTHER_ALL;
                        };
                        break;

                    /* reset to initial state */
                    case ',' :
                        who_u =
                        who_g =
                        who_o = 0;
                        grant = 0;
                        state = Q_EXPECT_WHO;
                        break;

                    /* trailing spaces section begins */
                    case ' ' :
                    case '\t' :
                        who_u =
                        who_g =
                        who_o = 0;
                        grant = 0;
                        state = Q_EXPECT_END;
                        break;

                    default :   /* unexpected char */
                        state = Q_STOP;
                        break;
                };
                break;

            /* eat trailing spaces */
            case Q_EXPECT_END :
                switch(*p_input){
                    case ' ' :  /* keep eating */
                    case '\t' :
                        break;

                    case ',' :  /* those spaces was not actually trailing? */
                        state = Q_EXPECT_WHO;   /* switch back to initial */
                        break;

                    default :   /* unexpected char */
                        state = Q_STOP;
                        break;
                };
                break;

            default :   /* to supress warning about unhandled enum values */
                state = Q_STOP;
                break;
        };
    }; 

    *p_error |= (state == Q_STOP) || (! (state & Q_FINI));

    return p_result;

#endif
}

row_acl
row_acl_from_sym_str(row_acl p_result, const char* p_input, int* p_error)
{
    return row_acl_from_sym_dat(p_result, p_input, -1, p_error);
}

row_acl
row_acl_from_dat(row_acl p_initial, const char* p_input, size_t p_size, int* p_error)
{
    /* should absolute permissions form allow leading/trailing
     * spaces too? */

    return  p_input[0] == '-' || p_input[0] == 'r' ?
                row_acl_from_abs_str(p_input, p_error)
            :
                row_acl_from_sym_dat(p_initial, p_input, p_size, p_error)
            ;
}

row_acl
row_acl_from_str(row_acl p_initial, const char* p_input, int* p_error)
{
    return row_acl_from_dat(p_initial, p_input, -1, p_error);
}
