/* -*- C -*-
 * File: "/home/arn/proj/cms1/db/row_acl/row_acl_text.h"
 * Author: "Anton Yabchinskiy, <arn@antora.ru>"
 * Copyright: ""
 * Created: "Fri, 13 Jul 2007 09:11:51 +0400 (arn@doriath.my.domain)"
 * Last modified: "Tue, 17 Jul 2007 14:36:53 +0400 (arn@doriath.my.domain)"
 * vim:set sts=4 ts=4 sw=4 ft=c:
 * $Id: row_acl_text.h,v 1.1 2007/07/17 10:59:22 anton Exp $
 */

#ifndef __ROW_ACL_TEXT_H_INCLUDED__
#define __ROW_ACL_TEXT_H_INCLUDED__	1

#include "row_acl.h"

/** Create text representation of _only_ one group of bits.
 * @param bits 5 bits of permissions
 * @param buffer output buffer (should be at least 5 + 1 octets long)
 * @return output buffer */
char*
row_acl_bits_to_str(row_acl, char*);

/** Create text representation of whole row_acl.
 * @param acl ACL to convert
 * @param buffer output buffer (should be at least ROW_ACL_STR_LENGTH
 * octets long)
 * @return output buffer */
char*
row_acl_to_str(row_acl, char*);

/** Create _group of bits_ from text representation.
 * @param input text representation
 * @param error error code will be stored here (caller is responsible
 * for this flag to be clear (have a value of 0))
 * @return result row acl. Returned value is undefined if error is set */
row_acl
row_acl_bits_from_str(const char*, int*);

/** Convert 'rw---r----r----' like string to row_acl
 * @param input text representation
 * @param error error code will be stored here (caller is responsible
 * for this flag to be clear (have a value of 0))
 * @return result row acl. Returned value is undefined if error is set */
row_acl
row_acl_from_abs_str(const char*, int*);

/** For unterminated strings
 */
row_acl
row_acl_from_sym_dat(row_acl, const char*, size_t, int*);

/** Convert 'u+rw,go+r' like string to row_acl
 * @param initial initial acl
 * @param input text representation
 * @param error error code will be stored here (caller is responsible
 * for this flag to be clear (have a value of 0))
 * @return result row acl. Returned value is undefined if error is set */
row_acl
row_acl_from_sym_str(row_acl, const char*, int*);

/** For unterminated strings
 */
row_acl
row_acl_from_dat(row_acl, const char*, size_t, int*);

/** Create row acl from it's text representation
 * @param initial initial acl for text in 'u-w,g-w,o-a'-like form
 * @param input text representation
 * @param error error code will be stored here (caller is responsible
 * for this flag to be clear (have a value of 0))
 * @return result row acl. Returned value is undefined if error is set */
row_acl
row_acl_from_str(row_acl, const char*, int*);

#endif/* ? __ROW_ACL_TEXT_H_INCLUDED__ */

